/* Copyright (C) 2004-2007, Murad Banaji
 *
 * This file is part of BRAINCIRC
 *
 * BRAINCIRC is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2, 
 * or (at your option) any later version.
 *
 * BRAINCIRC is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BRAINCIRC: see the file COPYING.  If not, write to 
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA. 

 * For comments, queries, or bug-reports about BRAINCIRC, please 
 * email braincirc@medphys.ucl.ac.uk

 */

#include "common/stringmanip.h"
#include "common/nrsimp.h"

#define min(A, B) ((A) < (B) ? (A) : (B))
#define max(A, B) ((A) > (B) ? (A) : (B))

int interpfile(double min_spac, double t_min, char *filename, char *fileoutname);

int genstepinputfile(char *inputparam, double initval, double stepval, char *outfilename, double timestep, double initdelay, double finaldelay, char smooth[], double smoothfactor){
  FILE *fd;
  int i;
  //va_list ap;
  char *p;
  if(strcmp(smooth, "smooth") != 0){
    fd = fopen(outfilename, "w");
    if (fd == NULL){
      fprintf(stderr, "ERROR in routine geninputfile in \"lininterp.c\": File %s could not be opened for writing.\n", outfilename);
      return -1;
    }
  }
  else{
    fd = fopen("models/inputs/__tempfile.dat", "w");
    if (fd == NULL){
      fprintf(stderr, "ERROR in routine geninputfile in \"lininterp.c\": \nFile \"models/inputs/__tempfile.dat\" could not be opened for writing.\n");
      return -1;
    }
  }

  fprintf(fd, "// This file is generated automatically. Do not edit or use directly\n\n", inputparam);
  fprintf(fd, "chosen_param: %s\n******\n", inputparam);
  if(timestep >0){
    if(initdelay>0)
      fprintf(fd, "%.4e, %.4e\n", initdelay, initval);
    fprintf(fd, "%.4e, %.4e\n", timestep, initval);
    fprintf(fd, "%.4e, %.4e\n", timestep, stepval);
    fprintf(fd, "%.4e, %.4e\n", timestep, initval);
    if(finaldelay>0)
      fprintf(fd, "%.4e, %.4e\n", finaldelay, initval);

  }

  if(timestep <=0){
    //  va_start(ap, tmp);
    fprintf(stderr, "ERROR in routine geninputfile in \"lininterp.c\":\n currently no facility for generating files with arbitrary inputs.\n");

  }
  fclose(fd);

 if(strcmp(smooth, "smooth") == 0){
   interpfile((initval-stepval)/smoothfactor, 1.0/smoothfactor, "models/inputs/__tempfile.dat", outfilename);
 }

  return 0;
}

int lininterp1(FILE *fd, double s, double t_min, double t2, double val1, double val2, int numparams){
  // s is the desired maximum step in the parameter value. It may not always be achievable unless t_min is sufficiently small 
  // t_min is the minimum time-step, to be used in the interpolation
  // t2 is the current time difference between the two values - the smoothing is done using time taken from the start of the second value's time
  // val1 is the starting value
  // val2 is the final value
  double tempval, tempspc;
  int num, i;
  if(numparams<1){
    fprintf(stderr, "ERROR in lininterp: number of parameters must be at least 1. Exiting.\n");
    return -1;
  }

  tempval = val2 - val1; // the current step in parameter value
  if(fabs(tempval) <= s) // nothing needs to be done - step already small enough
    fprintf(fd, "%.4e, %.4e\n", t2, val2);
  else{
    num = fabs(tempval)/s; // number of steps - uses integer division
    if((num+1)*t_min > t2){
      fprintf(fd, "%.4e, %.4e\n", t2, val2);
      fprintf(stderr, "WARNING in lininterp: time-step already too small.\n");
    }
    else{
      tempspc = tempval/(num+1);
      for(i=1;i<num+1;i++)
        fprintf(fd, "%.4e, %.4e\n", t_min, val1 + i*tempspc);
      fprintf(fd, "%.4e, %.4e\n", t2 - (num+1)*t_min, val2);
    }
  }
  return 0;
}




int  lininterp(FILE *fd, double s, double t_min, double t2, char line1[], char line2[]){
  // s is the desired maximum step in the parameter value. It may not always be achievable unless t_min is sufficiently small 
  // t_min is the minimum time-step, to be used in the interpolation
  // t2 is the current time difference between the two values - the smoothing is done using time taken from the start of the second value's time
  // line1 is the first line of data
  // line2 is the second line of data
  double tempval, tempspc;
  double *val1, *val2;
  int num, i, j;
  int numparams;
  char *word;

  numparams = getnumtokens(line1)-1;

  //  fprintf(stderr, "numparams = %d\n", numparams);

  if(getnumtokens(line1) != getnumtokens(line2)){
    fprintf(stderr, "ERROR in lininterp: number of parameters must be the same on both lines. Exiting.\n");
    fprintf(stderr, "%s\n%s", line1, line2);
    return -1;
  }
  // allocate memory

  val1 = dvector(0, numparams);
  val2 = dvector(0, numparams);

  // get in the parameter values

  for(j=0;j<numparams;j++){
    word = getnthtkn(line1, j+2);
    val1[j] = atof(word);
    free(word);
  }
  for(j=0;j<numparams;j++){
    word = getnthtkn(line2, j+2);
    val2[j] = atof(word);
    free(word);
  }

  tempval = val2[0] - val1[0]; // the current step in parameter value
  if(fabs(tempval) <= s){ // nothing needs to be done - parameter step already small enough
    fprintf(fd, "%s", line2);
  }
  else{
    num = fabs(tempval)/s; // number of steps - uses integer division
    if((num+1)*t_min > t2){
      fprintf(fd, "%s", line2);
      fprintf(stderr, "WARNING in lininterp: time-step already too small.\n");
    }
    else{
      for(i=1;i<num+1;i++){
        fprintf(fd, "%.4e", t_min);
	for(j=0;j<numparams;j++){
	  tempspc = (val2[j] - val1[j])/(num+1);
	  fprintf(fd, ", %.4e", val1[j] + i*tempspc);
	}
	fprintf(fd, "\n");
      }
      // last line
      fprintf(fd, "%.4e", t2 - (num+1)*t_min);
      for(j=0;j<numparams;j++)
	fprintf(fd, ", %.4e", val2[j]);
      fprintf(fd, "\n");
    }
  }
  // free memory
  free_dvector(val1, 0, numparams);
  free_dvector(val2, 0, numparams);
  return 0;
}





int interpfile(double min_spac, double t_min, char *filename, char *fileoutname){
  FILE *fd, *fd1;
  char block[10000];
  char line[1000];
  char line1[1000];
  char *pnames;
  char *val1;
  char **v2;
  int len, j, linenum=0;
  int numparams;
  char tempstr[30];
  double tstep;
  int block_pos;
  int numchange;
  double val=0;
  double vallast=0;
  int toknum;
  char *word;
  char *tempfname = "tempfiles/__temp_infile.dat";

  if(strcmp(filename, fileoutname) == 0){
    fprintf(stderr, "ERROR in routine interpfile in \"lininterp.c\": The input and output file names must be different.\n");
    return -1;
  }

  // first convert all input files to files with printed time-steps.

  infileconvert(filename, tempfname);

  fd = fopen(tempfname, "r");
  if (fd == NULL){
    fprintf(stderr, "ERROR in routine interpfile in \"lininterp.c\": File %s could not be opened for reading.\n", tempfname);
    return -1; //exit: datafile not found
  }
  fd1 = fopen(fileoutname, "w");
  if (fd1 == NULL){
    fprintf(stderr, "ERROR in routine interpfile in \"lininterp.c\": File %s could not be opened for writing.\n", fileoutname);
    fclose(fd);
    return -1; //exit: datafile not found
  }

  fprintf(fd1, "// This file is generated automatically. Do not edit or use directly.\n\n");

  len = getblock(fd, block, "******", 10000);
  block_pos = 0;
  while(block_pos <len){
    getlinefromblock(&block_pos, block, line, 1000);
    if(!(iscomsep(line)) && strstr(line, "chosen_param:") !=NULL){
      pnames = getrst("chosen_param:", line);
      numchange = getnumtokens(pnames);
      fprintf(fd1, "%s", line);
      free(pnames);
    }
    else
      fprintf(fd1, "%s", line);
  }

  linenum=0;
  if(t_min<=0){ // negative t_min means algorithm chooses timestep value
    while((len = getline(fd, line, 1000)) > 0){
      if(!(iscomsep(line))){
	word = getnthtkn(line, 1);
	if(linenum==0)
	  val=atof(word);
	else
	  val=min(val, atof(word));
	linenum++;
	free(word);
      }
    }
    t_min = min(0.5, val/1000.0); //0.5 or 0.1 % of the smallest value, whichever is smaller 
  }
  fclose(fd);

  // get the file pointer back to where it was
  fd=fopen(tempfname, "r");
  if (fd == NULL){
    fprintf(stderr, "ERROR in routine interpfile in \"lininterp.c\": File %s could not be opened for reading.\n", tempfname);
    return -1; //exit: datafile not found
  }
  len = getblock(fd, block, "******", 10000);

  linenum=0;
  if(min_spac<=0){ // negative min_spac means algorithm chooses step-value
    toknum=2;
    while((len = getline(fd, line, 1000)) > 0){
      if(!(iscomsep(line))){
	word = getnthtkn(line, toknum);
	if(linenum==0)
	  val=fabs(atof(word));
	else
	  val=max(val, fabs(atof(word)));
	linenum++;
	free(word);
      }
    }
    min_spac = val/(-min_spac); //e.g. if min_spac = -200, then we get 0.5% of the largest value 
    //      fprintf(stderr, "min_spac = %.4e\n", min_spac);

    fclose(fd);

    // get the file pointer back to where it was

    fd=fopen(tempfname, "r");
    if (fd == NULL){
      fprintf(stderr, "ERROR in routine interpfile in \"lininterp.c\": File %s could not be opened for reading.\n", tempfname);
      return -1; //exit: datafile not found
    }
    len = getblock(fd, block, "******", 10000);

  }

  v2 = charar(0, numchange+1);

  linenum=0;
  while((len = getline(fd, line, 1000)) > 0){
    if(iscomsep(line))
      fprintf(fd1, "%s", line);
    else{
      linenum++;
      numparams=chemgets(line, v2, numchange+2); // need for the time-step
      if(linenum == 1){ // first line
	fprintf(fd1, "%s", line);
      }
      else{ // subsequent lines
	if(lininterp(fd1, min_spac, t_min, atof(v2[0]), line1, line)!=0){
	  fclose(fd); fclose(fd1);
	  free_charar(v2, 0, numchange+1);
	  fprintf(stderr, "ERROR in routine interpfile in \"lininterp.c\": Exiting.\n", tempfname); 
	  return -1;
	}
      }
    }
    strcpy(line1, line); // safe
  }
  fclose(fd);
  fclose(fd1);
  free_charar(v2, 0, numchange+1);
  return 0;


}

int main(int argc, char *argv[])
{
  int i;
  if(argc !=5){
    fprintf(stderr, "ERROR in lininterp: must have the minimum parameter-step, minimum time-step and then two filenames as arguments.\n");
    return -1;
  }
  if(interpfile(atof(argv[1]), atof(argv[2]), argv[3], argv[4])!=0){
    fprintf(stderr, "Smoothing of file %s failed.", argv[3]);
    return -1;
  }
  return 0;
}
